package ru.dbotthepony.mc.otm.block.entity

import com.google.common.collect.ImmutableSet
import net.minecraft.world.level.block.entity.BlockEntityType
import net.minecraft.core.BlockPos
import net.minecraft.world.level.block.state.BlockState
import net.minecraft.world.MenuProvider
import net.minecraft.world.entity.player.Inventory
import net.minecraft.world.entity.player.Player
import net.minecraft.world.inventory.AbstractContainerMenu
import net.minecraft.nbt.CompoundTag
import net.minecraft.network.chat.Component
import net.minecraft.world.item.ItemStack
import net.minecraftforge.common.capabilities.ForgeCapabilities
import net.minecraftforge.items.IItemHandler
import ru.dbotthepony.mc.otm.capability.CombinedItemHandler
import ru.dbotthepony.mc.otm.capability.EmptyItemHandler
import ru.dbotthepony.mc.otm.capability.UnmodifiableItemHandler
import ru.dbotthepony.mc.otm.capability.moveBetweenSlots
import ru.dbotthepony.mc.otm.core.TextComponent
import ru.dbotthepony.mc.otm.core.getValue
import ru.dbotthepony.mc.otm.core.ifPresentK
import ru.dbotthepony.mc.otm.core.immutableMap
import ru.dbotthepony.mc.otm.core.math.RelativeSide
import ru.dbotthepony.mc.otm.core.nbt.getJson
import ru.dbotthepony.mc.otm.core.nbt.putJson
import ru.dbotthepony.mc.otm.core.nbt.set
import ru.dbotthepony.mc.otm.core.util.ITickable

/**
 * Device block entity base, implementing [MenuProvider] and [IRedstoneControlled], and also tracks custom display name
 */
abstract class MatteryDeviceBlockEntity(blockEntityType: BlockEntityType<*>, blockPos: BlockPos, blockState: BlockState)
	: MatteryBlockEntity(blockEntityType, blockPos, blockState), MenuProvider, IRedstoneControlled {
	var customDisplayName: Component? = null

	override val redstoneControl: AbstractRedstoneControl = RedstoneControl { new, old ->
		setChangedLight()

		if (new != old)
			redstoneStatusUpdated(new, old)
	}

	protected open val defaultDisplayName: Component
		get() = level?.getBlockState(blockPos)?.block?.name ?: TextComponent("null at $blockPos")

	abstract override fun createMenu(containerID: Int, inventory: Inventory, ply: Player): AbstractContainerMenu?

	override fun getDisplayName(): Component {
		return customDisplayName ?: defaultDisplayName
	}

	protected open fun redstoneStatusUpdated(newBlocked: Boolean, oldBlocked: Boolean) {}

	override fun saveAdditional(nbt: CompoundTag) {
		super.saveAdditional(nbt)

		if (customDisplayName != null)
			nbt.putJson("Name", Component.Serializer.toJsonTree(customDisplayName!!))

		nbt[REDSTONE_CONTROL_KEY] = redstoneControl.serializeNBT()
	}

	override fun load(nbt: CompoundTag) {
		super.load(nbt)

		customDisplayName = nbt.getJson("Name")?.let(Component.Serializer::fromJson)
		redstoneControl.deserializeNBT(nbt[REDSTONE_CONTROL_KEY] as? CompoundTag)
	}

	enum class ItemHandlerMode(val translationKey: String) {
		DISABLED("otm.gui.side_mode.disabled"),
		INPUT("otm.gui.side_mode.input"),
		OUTPUT("otm.gui.side_mode.output"),
		INPUT_OUTPUT("otm.gui.side_mode.input_output"),
		BATTERY("otm.gui.side_mode.battery");
	}

	inner class ConfigurableItemHandler(
		input: IItemHandler? = null,
		output: IItemHandler? = null,
		battery: IItemHandler? = null,
		val frontDefault: ItemHandlerMode = determineDefaultMode(input, output, battery, RelativeSide.FRONT),
		val backDefault: ItemHandlerMode = determineDefaultMode(input, output, battery, RelativeSide.BACK),
		val leftDefault: ItemHandlerMode = determineDefaultMode(input, output, battery, RelativeSide.LEFT),
		val rightDefault: ItemHandlerMode = determineDefaultMode(input, output, battery, RelativeSide.RIGHT),
		val topDefault: ItemHandlerMode = determineDefaultMode(input, output, battery, RelativeSide.TOP),
		val bottomDefault: ItemHandlerMode = determineDefaultMode(input, output, battery, RelativeSide.BOTTOM),
	) {
		val sideless: IItemHandler

		init {
			val caps = ArrayList<IItemHandler>()

			if (input != null) caps.add(input)
			if (output != null) caps.add(output)
			if (battery != null) caps.add(battery)

			sideless = UnmodifiableItemHandler(CombinedItemHandler(caps))
			exposeSideless(ForgeCapabilities.ITEM_HANDLER, sideless)
		}

		val front = Piece(RelativeSide.FRONT, input, output, battery).also { it.mode = frontDefault }
		val back = Piece(RelativeSide.BACK, input, output, battery).also { it.mode = backDefault }
		val left = Piece(RelativeSide.LEFT, input, output, battery).also { it.mode = leftDefault }
		val right = Piece(RelativeSide.RIGHT, input, output, battery).also { it.mode = rightDefault }
		val top = Piece(RelativeSide.TOP, input, output, battery).also { it.mode = topDefault }
		val bottom = Piece(RelativeSide.BOTTOM, input, output, battery).also { it.mode = bottomDefault }

		val sides = immutableMap {
			put(RelativeSide.FRONT, front)
			put(RelativeSide.BACK, back)
			put(RelativeSide.LEFT, left)
			put(RelativeSide.RIGHT, right)
			put(RelativeSide.TOP, top)
			put(RelativeSide.BOTTOM, bottom)
		}

		val defaults = immutableMap {
			put(RelativeSide.FRONT, frontDefault)
			put(RelativeSide.BACK, backDefault)
			put(RelativeSide.LEFT, leftDefault)
			put(RelativeSide.RIGHT, rightDefault)
			put(RelativeSide.TOP, topDefault)
			put(RelativeSide.BOTTOM, bottomDefault)
		}

		inner class Piece(
			side: RelativeSide,
			val input: IItemHandler? = null,
			val output: IItemHandler? = null,
			val battery: IItemHandler? = null,
		) : IItemHandler, ITickable {
			private var currentHandler: IItemHandler = EmptyItemHandler
			val capController = this@MatteryDeviceBlockEntity.sides[side]!!.Cap(ForgeCapabilities.ITEM_HANDLER, this)
			private val neighbour by this@MatteryDeviceBlockEntity.sides[side]!!.track(ForgeCapabilities.ITEM_HANDLER)

			val possibleViews: ImmutableSet<ItemHandlerMode>
			val inputOutput: IItemHandler?

			init {
				tickList.always(this)

				val builder = ImmutableSet.Builder<ItemHandlerMode>()

				builder.add(ItemHandlerMode.DISABLED)

				if (input != null) builder.add(ItemHandlerMode.INPUT)
				if (output != null) builder.add(ItemHandlerMode.OUTPUT)
				if (input != null && output != null) builder.add(ItemHandlerMode.INPUT_OUTPUT)
				if (battery != null) builder.add(ItemHandlerMode.BATTERY)

				possibleViews = builder.build()

				capController.close()

				if (input != null && output != null) {
					inputOutput = CombinedItemHandler(input, output)
				} else {
					inputOutput = null
				}
			}

			var mode by synchronizer.enum(ItemHandlerMode.DISABLED, setter = { value, access, setByRemote ->
				require(value in possibleViews) { "View type $value is not allowed (allowed views: $possibleViews)" }

				if (access.read() != value) {
					access.write(value)

					if (value == ItemHandlerMode.DISABLED) {
						capController.close()
					} else {
						capController.close()
						capController.expose()
					}

					when (value) {
						ItemHandlerMode.DISABLED -> currentHandler = EmptyItemHandler
						ItemHandlerMode.INPUT -> currentHandler = input!!
						ItemHandlerMode.OUTPUT -> currentHandler = output!!
						ItemHandlerMode.INPUT_OUTPUT -> currentHandler = inputOutput!!
						ItemHandlerMode.BATTERY -> currentHandler = battery!!
					}
				}
			})

			var automatePull = false
				set(value) {
					if (field != value) {
						field = value

						if (value) {
							innerSlotPull = 0
							outerSlotPull = 0
						}
					}
				}

			var automatePush = false
				set(value) {
					if (field != value) {
						field = value

						if (value) {
							innerSlotPush = 0
							outerSlotPush = 0
						}
					}
				}

			init {
				savetables.bool(::automatePull, "itemhandler_${side}_automatePull")
				savetables.bool(::automatePush, "itemhandler_${side}_automatePush")
				savetables.enum(::mode, "itemhandler_${side}_mode", ItemHandlerMode::valueOf)
			}

			private var innerSlotPull = 0
			private var outerSlotPull = 0

			private var innerSlotPush = 0
			private var outerSlotPush = 0

			override fun tick() {
				if (mode == ItemHandlerMode.DISABLED || currentHandler.slots == 0 || redstoneControl.isBlockedByRedstone)
					return

				neighbour.ifPresentK {
					if (it.slots == 0)
						return

					if (automatePull) {
						if (innerSlotPull !in 0 until currentHandler.slots)
							innerSlotPull = 0

						if (outerSlotPull !in 0 until it.slots)
							outerSlotPull = 0

						val (outerSlotPull, innerSlotPull) = moveBetweenSlots(it, outerSlotPull, currentHandler, innerSlotPull)
						this.innerSlotPull = innerSlotPull
						this.outerSlotPull = outerSlotPull
					}

					if (automatePush) {
						if (innerSlotPush !in 0 until currentHandler.slots)
							innerSlotPush = 0

						if (outerSlotPush !in 0 until it.slots)
							outerSlotPush = 0

						val (innerSlotPush, outerSlotPush) = moveBetweenSlots(currentHandler, innerSlotPush, it, outerSlotPush)
						this.innerSlotPush = innerSlotPush
						this.outerSlotPush = outerSlotPush
					}
				}
			}

			override fun getSlots(): Int {
				return currentHandler.slots
			}

			override fun getStackInSlot(slot: Int): ItemStack {
				return currentHandler.getStackInSlot(slot)
			}

			override fun insertItem(slot: Int, stack: ItemStack, simulate: Boolean): ItemStack {
				return currentHandler.insertItem(slot, stack, simulate)
			}

			override fun extractItem(slot: Int, amount: Int, simulate: Boolean): ItemStack {
				return currentHandler.extractItem(slot, amount, simulate)
			}

			override fun getSlotLimit(slot: Int): Int {
				return currentHandler.getSlotLimit(slot)
			}

			override fun isItemValid(slot: Int, stack: ItemStack): Boolean {
				return currentHandler.isItemValid(slot, stack)
			}
		}
	}

	companion object {
		private fun determineDefaultMode(input: IItemHandler?, output: IItemHandler?, battery: IItemHandler?, side: RelativeSide): ItemHandlerMode {
			if (side == RelativeSide.BACK && battery != null) {
				return ItemHandlerMode.BATTERY
			}

			if (input == null && output == null) {
				return ItemHandlerMode.DISABLED
			} else if (input != null && output != null) {
				return when (side) {
					RelativeSide.FRONT, RelativeSide.BACK -> ItemHandlerMode.DISABLED
					RelativeSide.RIGHT, RelativeSide.TOP -> ItemHandlerMode.INPUT
					RelativeSide.LEFT, RelativeSide.BOTTOM -> ItemHandlerMode.OUTPUT
				}
			} else if (output != null) {
				if (side == RelativeSide.TOP) {
					return ItemHandlerMode.DISABLED
				} else {
					return ItemHandlerMode.OUTPUT
				}
			} else {
				if (side == RelativeSide.BOTTOM) {
					return ItemHandlerMode.DISABLED
				} else {
					return ItemHandlerMode.INPUT
				}
			}
		}
	}
}
